<cfscript>
/**
 * Component that implements the prototype object model.
 *
 * Extend this component to create your own prototype, or create an instance
 * and build a chain at runtime. A component with optimize="true" 
 * 
 * @output false
 */
component
{
	/**
	 * @output false
	 */
	public any function init()
	{
		this.prototypeOptimized = false;
		
		return this;
	}
	
	/**
	 * Constructor for protoype instances.
	 * 
	 * @output false
	 */
	public any function constructor()
	{
	}
		
	/**
	 * @output false
	 */
	public any function new()
	{
		var instance = createObject(getMetaData(this).name).init(argumentCollection=arguments);
		
		if( structKeyExists(this,"prototype") )
			instance.prototype = this.prototype;
			
		instance.executeConstructors();
		
		return instance;
	}
		
	/**
	 * Proxies methods to the prototype and creates invokeProxies for methods that exist.
	 * 
	 * @output false
	 */
	public any function onMissingMethod(required string missingMethodName, required struct missingMethodArguments)
	{
		var method = getMethod(missingMethodName);
		
		if( this.prototypeOptimized or (structKeyExists(getMetaData(this),"optimized") and getMetaData(this).optimized) )
		{
			this[missingMethodName] = method;
		}
		else
		{
			this[missingMethodName] = invokeProxy;
		}
		
		return method(argumentCollection=missingMethodArguments);
	}
	
	/**
	 * Gets a function pointer by looking up the prototype chain.
	 *
	 * @output false
	 */
	public any function getMethod(required string name)
	{
		if( structKeyExists(this,name) and isCustomFunction(this[name]) )
			return this[name];
		
		if( structKeyExists(this,"prototype") )
			return this.prototype.getMethod(arguments.name);
			
		throw(type="MissingMethod", message="Method #arguments.name# doesn't exist.");
	}
	
	/**
	 * Executes each constructor.
	 *
	 * @output false
	 */
	public void function executeConstructors()
	{
		constructor(argumentCollection=arguments);
		
		for( var proto = this; structKeyExists(proto,"prototype"); )
		{
			proto = proto.prototype;
			var constructor = proto.constructor;
			constructor(argumentCollection=arguments);
		}
	}
	
	/**
	 * Invokes a method by name farther up the chain.
	 *
	 * @output false
	 */
	private any function invokePrototype(required string name, struct args={})
	{
		var method = this.prototype.getMethod(name);
		return method(argumentCollection=args);
	}
	
	/**
	 * Proxy to make invoked methods faster without usage of onMissingMethod().
	 * 
	 * @output false
	 */
	private any function invokeProxy()
	{
		var method = this.prototype.getMethod(getFunctionCalledName());
		return method(argumentCollection=arguments);
	}
}
</cfscript>